#!/usr/bin/python
#################################################################################################
# Synticka.py           A python interface to the Drum loadable kernel module for polyhponic
#                       sound synthesis on the RaspBerry 3. The raspberry should have
#                       ADCs and DACs connected by SPI, see tietonoita.fi. The kernel
#                       module is accessed trough the /proc/drum/ files.
#
# License               GPL                                             

from math import *
import os
import struct                                   # See https://docs.python.org/3.4/library/struct.html
import copy
import aifc
import pyqtgraph as pg

Home = os.path.expanduser('~')

#################################################################################################
# Commands
#
# MakeSig                   Sends a sound to the sound table, using the /proc/drum/Command file.
#
#                           u8          "A"
#                           u8          Sound number
#                           u32         amplitude
#                           u32         SustainStart
#                           u32         DecayStart
#                           u32         length
#                           u32 x n     data

def MakeSig(SoundIndex, LeftChannel, RightChannel, Amplitude=65535):
    if len(LeftChannel) != len(RightChannel):
        print "Erreur MakeSig: len(LeftChannel) != len(RightChannel)"
        return

    CommandWriteFile = open("/proc/drum/Command", "wb")
    CommandWriteFile.write("A" + struct.pack('<iiiii', SoundIndex, Amplitude, 0, 0, len(LeftChannel)) + "\x00\x00\x00") # Zeros for paddning to n x 4 bytes

    for Index in range(len(LeftChannel)):
        #print '<hh', LeftChannel[Index], RightChannel[Index]
        CommandWriteFile.write(struct.pack('<hh', LeftChannel[Index], RightChannel[Index]))

    CommandWriteFile.close()

def ReadSig(SoundIndex, SigDataIndex=0, SigDataIndexEnd=2000000000, ChIndex=-1):
    CommandWriteFile = open("/proc/drum/Command", "wb")
    CommandWriteFile.write("R" + struct.pack('<iii', SoundIndex, SigDataIndex, SigDataIndexEnd) + "\x00\x00\x00") # Zeros for paddning to n x 4 bytes
    CommandWriteFile.close()

    CommandReadFile = open("/proc/drum/Command", "rb")
    Signal = True
    SigList = []

    while (Signal):
        Signal = CommandReadFile.read()
        print "len(Signal): ", len(Signal)

        for i in range(len(Signal) / 4):
            #print i, struct.unpack_from('<hh', Signal, i*4)
            if ChIndex < 0 or ChIndex > 1: SigList.append(struct.unpack_from('<Hh', Signal, i*4))
            else:                          SigList.append(struct.unpack_from('<hh', Signal, i*4)[ChIndex])

    CommandReadFile.close()
    return SigList

def ReadSig32(SoundIndex, SigDataIndex=0, SigDataIndexEnd=2000000000):
    Array = ReadSig(SoundIndex, SigDataIndex, SigDataIndexEnd)

    for i in range(len(Array)):
        Array[i] = Array[i][0] + Array[i][1] * 0x10000

    return Array

def InitSig(SoundIndex, Size, Amplitude=65535):
    CommandWriteFile = open("/proc/drum/Command", "wb")
    CommandWriteFile.write("Z" + struct.pack('<iiiii', SoundIndex, Amplitude, 0, 0, Size) + "\x00\x00\x00") # Zeros for paddning to n x 4 bytes
    CommandWriteFile.close()

def SetSigPars(SoundIndex, Amplitude=65535):
    CommandWriteFile = open("/proc/drum/Command", "wb")
    CommandWriteFile.write("P" + struct.pack('<iiii', SoundIndex, Amplitude, 0, 0) + "\x00\x00\x00") # Zeros for paddning to n x 4 bytes
    CommandWriteFile.close()

def SetADCpars(Channel, SigIndex=-1, TrigLevel=10000, FilterFact=100, FilterDerivativeGap=9):
    CommandWriteFile = open("/proc/drum/Command", "wb")
    CommandWriteFile.write("F" + struct.pack('<iiiii', Channel, SigIndex, TrigLevel, FilterFact, FilterDerivativeGap) + "\x00\x00\x00") # Zeros for paddning to n x 4 bytes
    CommandWriteFile.close()

def Stop():
    CommandWriteFile = open("/proc/drum/Command", "wb")
    CommandWriteFile.write("S")
    CommandWriteFile.close()

# Audio file handling commands

def ToSigned(Low, High):
    if (Low + High*256) > 0x7FFF: return (Low + High*256) - 0x10000
    else:             return Low + High*256

def ReadAudioFile(Name):
    Sound = aifc.open(Name, "rb")

    if Sound.getnchannels() != 2:
        print "Erreur: audio file has less or more than two channels."
        return None

    if Sound.getsampwidth() != 2:
        print "Erreur: audio file is not 16 bit values per sample."

    print "Audio file ", Name, " has sample rate: ", Sound.getframerate(), " Hz, and is ", Sound.getnframes(), "*4 bytes long."

    Frames = Sound.readframes(Sound.getnframes())
    lCh = []
    rCh = []

    for i in range(0, Sound.getnframes()*4, 4):
        lCh.append(ToSigned(ord(Frames[i+1]), ord(Frames[i])))
        rCh.append(ToSigned(ord(Frames[i+3]), ord(Frames[i+2])))

    return lCh, rCh

# Debug and simulation commands

def Input(FromInd, Index):
    global Data
    if Index < 0 or Index > len(Data[FromInd]): return 0.0
    else:                                       return Data[FromInd][Index]

def Output(ToInd, Index):
    global Data
    if Index < 0 or Index > len(Data[ToInd]): return 0.0
    else:                                     return Data[ToInd][Index]

def FilterStep(Index, FromInd, ToInd):
    Sum = 0

    for i in range(len(InCoeff)):
        Sum += Input(FromInd, Index-i) * InCoeff[i] + Output(ToInd, Index-i) * OutCoeff[i]

    return Sum

def Filter(FromInd, ToInd):
    Data[ToInd] = copy.copy(Data[FromInd])

    for Index in range(len(Data[FromInd])):
        Data[ToInd][Index] = FilterStep(Index, FromInd, ToInd)

def sFilterStep(Index, FromInd, ToInd):
    Sum = 0

    for i in range(len(InCoeff)):
        if Input(FromInd, Index-i) * InCoeffPos[i] + Output(ToInd, Index-i) * OutCoeffPos[i] > \
           Input(FromInd, Index-i) * InCoeffNeg[i] + Output(ToInd, Index-i) * OutCoeffNeg[i]:
              Sum += Input(FromInd, Index-i) * InCoeffPos[i] + Output(ToInd, Index-i) * OutCoeffPos[i]
        else: Sum += Input(FromInd, Index-i) * InCoeffNeg[i] + Output(ToInd, Index-i) * OutCoeffNeg[i]

    return Sum

def sFilter(FromInd, ToInd):
    Data[ToInd] = copy.copy(Data[FromInd])

    for Index in range(len(Data[FromInd])):
        Data[ToInd][Index] = sFilterStep(Index, FromInd, ToInd)

def Abs(SetNo):
    for Index in range(len(Data[SetNo])):
        Data[SetNo][Index] = abs(Data[SetNo][Index])

def Pow(SetNo, Exp):
    for Index in range(len(Data[SetNo])):
        Data[SetNo][Index] = pow(Data[SetNo][Index], Exp)

def Fact(SetNo, Fact):
    for Index in range(len(Data[SetNo])):
        Data[SetNo][Index] = Data[SetNo][Index] * Fact

def Offset(SetNo, Offset):
    for Index in range(len(Data[SetNo])):
        Data[SetNo][Index] = Data[SetNo][Index] + Offset

def Comp(SetNo, CompVal, HighVal=1000, LowVal=-1000):
    for Index in range(len(Data[SetNo])):
        if Data[SetNo][Index] > CompVal: Data[SetNo][Index] = HighVal
        else:                            Data[SetNo][Index] = LowVal

def SaveSignal(Array, FileName):
    File = open(FileName, "w")
    File.write("DataSet = ")
    File.write(str(Array))
    File.close()

def Copy(FromInd, ToInd):
    Data[ToInd] = copy.copy(Data[FromInd])

def Plot(SetNo):
    Graph[SetNo] = PlotWidget.plot(Data[SetNo], pen=(SetNo, 9))

def Clear(SetNo=None):
    if SetNo == None: PlotWidget.clear()
    else:             Graph[SetNo].clear()
 
PlotWidget = pg.plot(title="Drum ADC sig filtering")

InCoeff  = [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]  # Coeff for newest data first...
OutCoeff = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]  # OBS OutCoeff[0] verkar paa samma
                                                               # data som InCoeff[0]...
InCoeffPos  = [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]  # Coeff for newest data first...
OutCoeffPos = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]  # OBS OutCoeff[0] verkar paa samma
                                                               # data som InCoeff[0]...
InCoeffNeg  = [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]  # Coeff for newest data first...
OutCoeffNeg = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]  # OBS OutCoeff[0] verkar paa samma
                                                               # data som InCoeff[0]...

Data = [None, None, None, None, None, None, None, None, None, None, \
        None, None, None, None, None, None, None, None, None, None]

Graph = [None, None, None, None, None, None, None, None, None, None, \
          None, None, None, None, None, None, None, None, None, None]

def Process(SetNo):
    global InCoeff, OutCoeff

    InCoeff  = [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0]
    OutCoeff = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    Filter(SetNo, SetNo+1)
    Pow(SetNo+1, 2)
    InCoeff  = [0.01, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    OutCoeff = [0.0, 0.99, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    Filter(SetNo+1, SetNo+2)
    Fact(SetNo+2, 0.001)
    Plot(SetNo)
    Offset(SetNo+2, 3000)
    Plot(SetNo+2)

def MakeSweep(fact=0.001, offs=0):
    Sweep = []
    for x in range(88200):
        SineWavePt = sin(pow(x+offs,2)*fact)
        Sweep.append(int(SineWavePt*20000))
    return Sweep

def MakeBong(Freq=441, Decay=0.0005):
    Sweep = []
    for x in range(44100):
        #SineWavePt = sin(x / 44100.0 * Freq * pi / 2.0) + \
        #             sin(x / 44100.0 * Freq*3 * pi / 2.0) * 0.2 + \
        #             sin(x / 44100.0 * Freq*5 * pi / 2.0) * 0.1
        SineWavePt = sin(x / 44100.0 * Freq * pi / 2.0) + \
                     sin(x / 44100.0 * Freq*2 * pi / 2.0) * 0.2 + \
                     sin(x / 44100.0 * Freq*6 * pi / 2.0) * 0.1
        Sweep.append(int(SineWavePt*20000 / (x*Decay+1)))
    return Sweep

PitchFact = 0.3

Data[0] = MakeBong(392.00 * PitchFact)
MakeSig(25, Data[0], Data[0])
SetADCpars(0, 25, TrigLevel=10000, FilterFact=60, FilterDerivativeGap=7)

Data[0] = MakeBong(440.00 * PitchFact)
MakeSig(26, Data[0], Data[0])
SetADCpars(1, 26, TrigLevel=10000, FilterFact=60, FilterDerivativeGap=7)

Data[0] = MakeBong(493.88 * PitchFact)
MakeSig(27, Data[0], Data[0])
SetADCpars(2, 27, TrigLevel=10000, FilterFact=60, FilterDerivativeGap=7)

Data[0] = MakeBong(523.25 * PitchFact)
MakeSig(28, Data[0], Data[0])
SetADCpars(3, 28, TrigLevel=10000, FilterFact=60, FilterDerivativeGap=7)

Data[0] = MakeBong(587.33 * PitchFact)
MakeSig(29, Data[0], Data[0])
SetADCpars(4, 29, TrigLevel=10000, FilterFact=60, FilterDerivativeGap=7)

Data[0] = MakeBong(659.25 * PitchFact)
MakeSig(30, Data[0], Data[0])
SetADCpars(5, 30, TrigLevel=10000, FilterFact=60, FilterDerivativeGap=7)  # False trigges !

Data[0] = MakeBong(698.46 * PitchFact)
MakeSig(31, Data[0], Data[0])
SetADCpars(6, 31, TrigLevel=10000, FilterFact=60, FilterDerivativeGap=7)

Data[0] = MakeBong(783.99 * PitchFact)
MakeSig(32, Data[0], Data[0])
SetADCpars(7, 32, TrigLevel=10000, FilterFact=60, FilterDerivativeGap=6)












